/*
** pork_buddy_list.c - Buddy list screen widget.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <pork.h>
#include <pork_slist.h>

static void blist_cell_destroy_cb(void *d) {
	struct slist_cell *cell = d;

	if (cell->type == TYPE_LIST_CELL) {
		struct group *group = cell->data;

		group->blist_line = NULL;
	} else {
		struct buddy *buddy= cell->data;

		buddy->blist_line = NULL;
	}
}

static void blist_make_indicator(struct buddy *buddy, char *buf, size_t len) {

	switch (buddy->status) {
		case STATUS_AWAY:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_AWAY), len);
			break;

		case STATUS_IDLE:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_IDLE), len);
			break;

		case STATUS_ACTIVE:
			xstrncpy(buf, opt_get_str(OPT_FORMAT_BUDDY_ACTIVE), len);
			break;

		default:
			xstrncpy(buf, "%p?%x", len);
			break;
	}

}

static void blist_make_idle(struct buddy *buddy, char *buf, size_t len) {
	u_int32_t i;
	u_int32_t j;
	u_int32_t s = 0;
	char *format = opt_get_str(OPT_FORMAT_BLIST_IDLE);
	char tmp[32];

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			if (format[++i] == 'I') {
				time_to_str(buddy->idle_time, tmp, sizeof(tmp));
				for (j = 0 ; tmp[j] != '\0' && s < len && j < sizeof(tmp) ; j++)
					buf[s++] = tmp[j];
			} else {
				buf[s++] = format[--i];
			}
		} else {
			buf[s++] = format[i];
		}
	}
	buf[s] = '\0';
}

static void blist_make_warn(struct buddy *buddy, char *buf, size_t len) {
	u_int32_t i;
	u_int32_t j;
	u_int32_t s = 0;
	char *format = opt_get_str(OPT_FORMAT_BLIST_WARN);
	char tmp[32];

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			if (format[++i] == 'W') {
				snprintf(tmp, sizeof(tmp), "%u", buddy->warn_level);
				for (j = 0 ; tmp[j] != '\0' && s < len && j < sizeof(tmp) ; j++)
					buf[s++] = tmp[j];
			} else {
				buf[s++] = format[--i];
			}
		} else {
			buf[s++] = format[i];
		}
	}
	buf[s] = '\0';
}

static void blist_make_buddy_label(	struct blist *blist __notused,
									struct buddy *buddy,
									chtype *ch,
									size_t len)
{
	u_int32_t i;
	u_int32_t s = 0;
	char tmp[32];
	char *format = opt_get_str(OPT_FORMAT_BLIST);
	char *label_left = xmalloc(len);
	char *label_right = NULL;
	char *label = label_left;

	len--;

	for (i = 0 ; format[i] != '\0' && s < len ; i++) {
		if (format[i] == FORMAT_VARIABLE) {
			*tmp = '\0';
			switch (format[++i]) {
				case 'B':
				case 'b':
					blist_make_indicator(buddy, tmp, sizeof(tmp));
					break;

				case 'N':
				case 'n':
					xstrncpy(tmp, buddy->name, sizeof(tmp));
					break;

				case 'I':
					blist_make_idle(buddy, tmp, sizeof(tmp));
					break;

				case 'i':
					if (buddy->idle_time > 0)
						blist_make_idle(buddy, tmp, sizeof(tmp));
					break;

				case 'W':
					blist_make_warn(buddy, tmp, sizeof(tmp));
					break;

				case 'w':
					if (buddy->warn_level > 0)
						blist_make_warn(buddy, tmp, sizeof(tmp));
					break;

				case '>':
					if (label_right == NULL) {
						label[s] = '\0';
						label_right = xmalloc(len + 1);
						label = label_right;
						s = 0;
						break;
					}

				default:
					label[s++] = format[--i];
					break;

			}

			if (*tmp != '\0') {
				u_int32_t j = 0;

				while (tmp[j] != '\0' && j < sizeof(tmp) && s < len)
					label[s++] = tmp[j++];
			}

		} else {
			label[s++] = format[i];
		}
	} 

	label[s] = '\0';

	if (label_right == NULL) {
		plaintext_to_cstr(label, ch, len);
		free(label);
	} else {
		size_t len_l;
		size_t len_r;
		chtype *chl;
		chtype *chr;
		u_int32_t i = 0;
		u_int32_t j = 0;
		u_int32_t fill_len;

		len_l = strlen(label_left) + 1;
		len_r = strlen(label_right) + 1;

		chl = xmalloc(sizeof(chtype) * len_l);
		chr = xmalloc(sizeof(chtype) * len_r);

		plaintext_to_cstr(label_left, chl, len_l);
		plaintext_to_cstr(label_right, chr, len_r);

		while (i < len && chl[i] != 0) {
			ch[i] = chl[i];
			i++;
		}

		fill_len = blist->slist.cols - i - cstrlen(chr) - 2;
		while (i < len && fill_len > 0) {
			ch[i++] = ' ';
			fill_len--;
		}

		while (i < len && chr[j] != 0)
			ch[i++] = chr[j++];

		ch[i] = 0;

		free(chl);
		free(chr);
		free(label_left);
		free(label_right);
	}
}

static void blist_make_group_label(	struct blist *blist __notused,
									struct group *group,
									chtype *ch,
									size_t len)
{
	char *label = xmalloc(len);
	char expand_char = '-';
	dlist_t *node = group->blist_line;

	if (node != NULL) {
		struct slist_cell *cell = node->data;

		if (cell != NULL && cell->collapsed)
			expand_char = '+';
	}

	snprintf(label, len, "%c %s (%u/%u)",
		expand_char, group->name, group->num_online,
		group->num_members);

	plaintext_to_cstr(label, ch, len);
	free(label);
}

void blist_update_label(struct blist *blist, dlist_t *node) {
	struct slist_cell *cell;

	if (blist == NULL || node == NULL)
		return;
	cell = node->data;

	if (cell->type == TYPE_FLAT_CELL) {
		struct buddy *buddy = cell->data;

		blist_make_buddy_label(blist, buddy, cell->label, blist->label_len);
	} else {
		struct group *group = cell->data;

		blist_make_group_label(blist, group, cell->label, blist->label_len);
	}

	if (cell->line >= 0) {
		if (node == blist->slist.cursor)
			blist_draw_cursor(blist, 1);
		else
			blist_draw_line(blist, cell);

		blist->slist.dirty = 1;
	}
}

int blist_init(struct aim_priv *priv) {
	int cols = BUDDY_DEFAULT_WIDTH;
	struct blist *blist;
	dlist_t *cur;
	int ret;

	if (priv == NULL || priv->blist != NULL)
		return (-1);

	blist = xcalloc(1, sizeof(*blist));

	ret = slist_init(&blist->slist, screen.rows - STATUS_ROWS, cols,
			screen.cols - cols, 0);
	if (ret != 0) {
		free(blist);
		return (-1);
	}

	slist_cell_free_cb(&blist->slist, blist_cell_destroy_cb);

	blist->label_len = 64;

	/* Add the buddy groups to the buddy list */
	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct group *group = cur->data;
		dlist_t *node = blist_add_group(blist, group);

		if (node != NULL)
			group->blist_line = node;

		cur = cur->next;
	}

	blist_draw(blist);
	priv->blist = blist;
	return (0);
}

dlist_t *blist_add_group(struct blist *blist, struct group *group) {
	struct slist_cell *cell;
	dlist_t *ret;

	if (blist == NULL)
		return (NULL);

	cell = xcalloc(1, sizeof(*cell));
	cell->type = TYPE_LIST_CELL;
	cell->parent = NULL;
	cell->children = NULL;
	cell->collapsed = 0;
	cell->refnum = group->refnum;
	cell->label = xmalloc(sizeof(chtype) * blist->label_len);
	cell->data = group;
	blist_make_group_label(blist, group, cell->label, blist->label_len);

	ret = slist_add(&blist->slist, cell);

	if (ret == NULL)
		return (NULL);

	group->blist_line = ret;
	blist_draw(blist);
	return (ret);
}

dlist_t *blist_add(struct blist *blist, struct buddy *buddy) {
	struct slist_cell *cell;
	dlist_t *ret;
	dlist_t *node;

	if (blist == NULL)
		return (NULL);

	cell = xcalloc(1, sizeof(*cell));
	cell->type = TYPE_FLAT_CELL;
	cell->parent = buddy->group->blist_line;
	cell->refnum = buddy->refnum;
	cell->label = xmalloc(sizeof(chtype) * blist->label_len);
	cell->data = buddy;
	blist_make_buddy_label(blist, buddy, cell->label, blist->label_len);

	ret = slist_add(&blist->slist, cell);
	if (ret == NULL)
		return (NULL);

	node = buddy->group->blist_line;
	cell = node->data;

	buddy->blist_line = ret;
	blist_make_group_label(blist, buddy->group, cell->label, blist->label_len);
	blist_draw(blist);
	return (ret);
}

void blist_del(struct blist *blist, struct buddy *buddy) {
	dlist_t *node;
	struct slist_cell *cell;

	if (blist == NULL)
		return;

	node = buddy->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;
	buddy->blist_line = NULL;
	slist_del(&blist->slist, cell);

	node = buddy->group->blist_line;
	cell = node->data;
	blist_make_group_label(blist, buddy->group, cell->label, blist->label_len);

	blist_draw(blist);
}

void blist_del_group(struct blist *blist, struct group *group) {
	dlist_t *node;
	struct slist_cell *cell;

	if (blist == NULL)
		return;

	node = group->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;
	group->blist_line = NULL;

	slist_del(&blist->slist, cell);
	blist_draw(blist);
}

void blist_collapse_group(struct blist *blist, struct group *group) {
	dlist_t *node;
	struct slist_cell *cell;

	node = group->blist_line;
	if (node == NULL)
		return;

	cell = node->data;
	if (cell == NULL)
		return;

	if (!cell->collapsed) {
		if (slist_collapse_list_cell(&blist->slist, group->blist_line) != 0)
			return;
	} else {
		dlist_t *cur = group->member_list;

		cell->collapsed = 0;

		while (cur != NULL) {
			struct buddy *buddy = cur->data;

			if (buddy->status != STATUS_OFFLINE)
				blist_add(blist, buddy);

			cur = cur->next;
		}
	}

	blist_make_group_label(blist, group, cell->label, blist->label_len);
	blist_draw(blist);
}

int blist_show(struct blist *blist, struct imwindow *imwindow) {
	u_int32_t new_width = imwindow->swindow.cols - blist->slist.cols;

	if (imwindow->blist_visible)
		return (-1);

	wclear(imwindow->swindow.win);
	swindow_resize(&imwindow->swindow, imwindow->swindow.rows, new_width);

	blist_draw(blist);

	imwindow->blist_visible = 1;
	imwindow_refresh(imwindow);

	return (0);
}

int blist_hide(struct blist *blist, struct imwindow *imwindow) {
	u_int32_t new_width = imwindow->swindow.cols + blist->slist.cols;

	if (!imwindow->blist_visible)
		return (-1);

	imwindow->active_binds = &screen.binds.main;
	imwindow->blist_visible = 0;

	wclear(imwindow->swindow.win);
	wnoutrefresh(imwindow->swindow.win);
	swindow_resize(&imwindow->swindow, imwindow->swindow.rows, new_width);
	imwindow_refresh(imwindow);

	return (0);
}

int blist_cursor_up(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_up(&blist->slist);
	blist_draw_cursor(blist, 1);

	if (ret == 0)
		blist->slist.dirty = 1;

	return (ret);
}

int blist_cursor_down(struct blist *blist) {
	int ret;

	if (blist == NULL)
		return (-1);

	blist_draw_cursor(blist, 0);
	ret = slist_cursor_down(&blist->slist);
	blist_draw_cursor(blist, 1);

	if (ret == 0)
		blist->slist.dirty = 1;

	return (ret);
}

void blist_draw_border(struct blist *blist) {
	struct imwindow *imwindow = screen.cur_window->data;
	attr_t box_attr;

	if (blist == NULL)
		return;

	if (imwindow->active_binds == &screen.binds.main)
		box_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_NOFOCUS));
	else
		box_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_FOCUS));

	slist_draw_border(&blist->slist, box_attr);
}

inline struct slist_cell *blist_get_cursor(struct blist *blist) {
	if (blist == NULL)
		return (NULL);

	return (slist_get_cursor(&blist->slist));
}

inline int blist_resize(struct blist *blist, u_int32_t rows, u_int32_t cols) {
	int ret;

	if (blist == NULL)
		return (0);

	ret = slist_resize(&blist->slist, rows, cols);
	if (ret != 0)
		return (-1);

	blist->label_len = 64;
	blist->slist.dirty = 1;

	return (0);
}

inline void blist_destroy(struct blist *blist) {
	if (blist == NULL)
		return;

	slist_destroy(&blist->slist);
	memset(blist, 0, sizeof(*blist));
}

inline void blist_draw(struct blist *blist) {
	if (blist == NULL)
		return;

	slist_draw(&blist->slist);
	wclrtobot(blist->slist.win);
	blist_draw_cursor(blist, 1);
	blist_draw_border(blist);
	blist->slist.dirty = 1;
}

inline void blist_draw_cursor(struct blist *blist, int status) {
	attr_t curs_attr = 0;

	if (status)
		curs_attr = r_cattr(opt_get_str(OPT_COLOR_BLIST_SELECTOR));

	slist_draw_cursor(&blist->slist, curs_attr);
}

inline void blist_draw_line(struct blist *blist, struct slist_cell *cell) {
	slist_draw_line(&blist->slist, cell);
}

inline void blist_refresh(struct blist *blist, struct imwindow *imwindow) {
	if (blist == NULL)
		return;

	if (imwindow->blist_visible)
		slist_refresh(&blist->slist);
}

// Called when you change a color setting for the buddy list

void blist_cch(void) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct blist *blist = imwindow->priv->blist;

	if (blist == NULL)
		return;

	blist->slist.dirty = 1;

	if (imwindow->blist_visible)
		blist_draw(blist);
}

// Called when you change a format string for the buddy list

void blist_newform(void) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct blist *blist = imwindow->priv->blist;
	dlist_t *p;

	if (blist == NULL)
		return;

	p = blist->slist.cells;
	while (p != NULL) {
		blist_update_label(blist, p);
		p = slist_node_next(p);
	}

	blist_cch();
}
