/*
** pork_screen.c - screen management.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <string.h>
#include <sys/types.h>

#include <pork.h>

/*
** Create the program's status window.  This window will have the buddy
** list visible in it by default.
*/

int screen_init(struct screen *scr, u_int32_t rows, u_int32_t cols) {
	struct imwindow *imwindow;
	struct aim_acct *acct;
	aim_session_t *session;
	aim_conn_t *input_conn;

	memset(scr, 0, sizeof(*scr));

	scr->rows = rows;
	scr->cols = cols;

	bind_init(&scr->binds);
	input_init(&scr->input, cols);

	if (status_init(scr) == -1) {
		input_destroy(&scr->input);
		return (-1);
	}

	memset(&scr->null_priv, 0, sizeof(scr->null_priv));
	scr->null_priv.username = xstrdup("<not specified>");
	aim_setup(&scr->null_priv);
	session = &scr->null_priv.aim_session;

	input_conn = aim_newconn(session, 0, NULL);
	if (input_conn == NULL)
		return (-1);

	input_conn->fd = STDIN_FILENO;

	acct = xcalloc(1, sizeof(*acct));
	acct->username = xstrdup(scr->null_priv.username);
	acct->session = session;
	acct->priv = &scr->null_priv;

	scr->acct_list = dlist_add_head(scr->acct_list, acct);

	imwindow = imwindow_new(rows, cols, 1, TYPE_STATUS,
		&scr->null_priv, "status");

	if (imwindow == NULL) {
		input_destroy(&scr->input);
		delwin(scr->status_bar);
		return (-1);
	}

	wopt_set(imwindow, WOPT_SHOW_BUDDY_LIST, "1");

	screen_add_window(scr, imwindow);

	scr->status_win = imwindow;
	return (0);
}

void screen_window_list_add(struct screen *scr, dlist_t *new_node) {
	struct imwindow *imwindow = new_node->data;

	/*
	** The window list is a sorted circular doubly linked list.
	** The window_list pointer points to the window having the lowest
	** refnum.
	*/

	if (scr->window_list == NULL) {
		new_node->prev = new_node;
		new_node->next = new_node;
		scr->window_list = new_node;
	} else {
		dlist_t *cur = scr->window_list;

		do {
			struct imwindow *imw = cur->data;

			if (imwindow->refnum < imw->refnum) {
				if (cur == scr->window_list)
					scr->window_list = new_node;
				break;
			}

			cur = cur->next;
		} while (cur != scr->window_list);

		new_node->next = cur;
		new_node->prev = cur->prev;

		new_node->next->prev = new_node;
		new_node->prev->next = new_node;
	}
}

void screen_window_list_remove(struct screen *scr, dlist_t *node) {
	dlist_t *save = node;

	if (node == scr->window_list)
		scr->window_list = node->next;

	if (node == scr->window_list)
		scr->window_list = NULL;

	save->prev->next = node->next;
	save->next->prev = node->prev;
}

void screen_add_window(struct screen *scr, struct imwindow *imwindow) {
	dlist_t *new_node = xmalloc(sizeof(*new_node));

	new_node->data = imwindow;

	screen_window_list_add(scr, new_node);

	/*
	** If this is the first window, make it current.
	*/

	if (scr->cur_window == NULL)
		screen_window_swap(scr, new_node);
}

/*
** Change the refnum on the currently visible window.
*/

int screen_renumber(struct screen *scr, u_int32_t refnum) {
	dlist_t *node;
	struct imwindow *imwindow;
	u_int32_t old_refnum;

	node = scr->cur_window;
	screen_window_list_remove(scr, node);

	imwindow = node->data;
	old_refnum = imwindow->refnum;
	imwindow->refnum = refnum;

	if (node != node->next || node->next != node->prev) {
		dlist_t *temp = screen_find_refnum(scr, refnum);
		if (temp != NULL) {
			struct imwindow *imw;

			screen_window_list_remove(scr, temp);
			imw = temp->data;
			imw->refnum = old_refnum;
			screen_window_list_add(scr, temp);
		}
	}

	screen_window_list_add(scr, node);
	return (0);
}

void screen_resize(struct screen *scr, u_int32_t rows, u_int32_t cols) {
	dlist_t *cur = scr->window_list;
	struct imwindow *imwindow = scr->cur_window->data;

	scr->rows = rows;
	scr->cols = cols;

	do {
		struct imwindow *win = cur->data;

		wclear(win->swindow.win);
		wnoutrefresh(win->swindow.win);
		imwindow_resize(cur->data, rows, cols);
		cur = cur->next;
	} while (scr->window_list != cur);

	mvwin(scr->status_bar, rows - STATUS_ROWS, 0);
	input_resize(&scr->input, cols);

	blist_resize(imwindow->priv->blist, rows - STATUS_ROWS, cols);
}

/*
** Find the window having the specified refnum, and return
** a pointer to the node that holds it.
**
** This depends on the list being sorted.
*/

dlist_t *screen_find_refnum(struct screen *scr, u_int32_t refnum) {
	dlist_t *cur = scr->window_list;

	do {
		struct imwindow *imwindow = cur->data;

		if (imwindow->refnum == refnum)
			return (cur);

		if (imwindow->refnum > refnum)
			break;

		cur = cur->next;
	} while (cur != scr->window_list);

	return (NULL);
}

/*
** Yes, this is pretty slow and stupid, but it'd be pretty
** unusual (i think) to have even 10 windows open at any one time.
**
** It's also only called when creating a new window.
*/

u_int32_t screen_get_new_refnum(struct screen *scr) {
	u_int32_t i;

	for (i = 1 ; i < ~0U ; i++) {
		if (screen_find_refnum(scr, i) == NULL)
			return (i);
	}

	return (0);
}

int screen_acct_find(void *l, void *r) {
	char *left = l;
	struct aim_acct *acct = r;

	return (strcasecmp(left, acct->username));
}

struct aim_acct *screen_add_acct(struct screen *scr, char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];
	struct aim_priv *priv;

	normalize(user, nname, sizeof(nname));

	node = dlist_find(scr->acct_list, nname, screen_acct_find);
	if (node != NULL)
		return (NULL);

	priv = pork_priv_init(nname);
	if (priv == NULL)
		return (NULL);

	acct = xcalloc(1, sizeof(*acct));
	acct->username = xstrdup(nname);
	acct->priv = priv;
	acct->session = &priv->aim_session;

	scr->acct_list = dlist_add_head(scr->acct_list, acct);
	return (acct);
}

int screen_del_acct(struct screen *scr, char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];

	normalize(user, nname, sizeof(nname));

	node = dlist_find(scr->acct_list, nname, screen_acct_find);
	if (node == NULL)
		return (-1);
	acct = node->data;

	scr->acct_list = dlist_remove(scr->acct_list, node);

	node = scr->window_list;
	if (node != NULL) {
		do {
			struct imwindow *win = node->data;

			if (win->priv == acct->priv) {
				if (win->blist_visible)
					blist_hide(win->priv->blist, win);

				win->priv->ref_count--;
				win->priv = &scr->null_priv;
				win->priv->ref_count++;
			}

			node = node->next;
		} while (node != scr->window_list);
	}

	/* This must always be the case. */
	if (acct->priv->ref_count == 0)
		pork_priv_free(acct->priv);

	free(acct->username);
	free(acct);

	return (0);
}
